"""
Template tags as workaround for a Django 1.11 bug in the multiwidget template.
"""
from django.template import Library, Node, TemplateSyntaxError

register = Library()


class WithDictNode(Node):
    """
    Node to push a complete dict
    """

    def __init__(self, nodelist, context_expr):
        self.nodelist = nodelist
        self.context_expr = context_expr

    def render(self, context):
        """
        Render the tag, with extra context layer.
        """
        extra_context = self.context_expr.resolve(context)
        if not isinstance(extra_context, dict):
            raise TemplateSyntaxError("{% withdict %} expects the argument to be a dictionary.")

        with context.push(**extra_context):
            return self.nodelist.render(context)


@register.tag
def withdict(parser, token):
    """
    Take a complete context dict as extra layer.
    """
    bits = token.split_contents()
    if len(bits) != 2:
        raise TemplateSyntaxError("{% withdict %} expects one argument")

    nodelist = parser.parse(('endwithdict',))
    parser.delete_first_token()

    return WithDictNode(
        nodelist=nodelist,
        context_expr=parser.compile_filter(bits[1])
    )



# following PEP 440
__version__ = "2.5.1"

# Values starting with such prefix should be handled as external URL.
EXTERNAL_SCHEMES = ('http', 'https', 'ftp', 'ftps', 'sftp', 'webdav', 'webdavs', 'afp', 'smb', 'git', 'svn', 'hg', 'mailto', 'tel')



import re

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

from any_urlfield import EXTERNAL_SCHEMES

try:
    from urllib.parse import urlparse  # Python 3
except ImportError:
    from urlparse import urlparse


class ExtendedURLValidator(URLValidator):
    """
    Add more schemes to those that will pass validation.
    """
    schemes = URLValidator.schemes + list(EXTERNAL_SCHEMES)

    # Phone numbers don't match the host regex in Django's validator,
    # so we test for a simple alternative.
    tel_re = r'^[0-9\#\*\-\.\(\)\+]+$'

    def __call__(self, value):
        try:
            super(ExtendedURLValidator, self).__call__(value)
        except ValidationError:
            parsed = urlparse(value)
            if parsed.scheme == "tel" and re.match(self.tel_re, parsed.netloc):
                return
            raise



from __future__ import unicode_literals

from unittest import skipIf

import django
from django.core.exceptions import ValidationError
from django.test import TestCase

from any_urlfield.validators import ExtendedURLValidator


class ValidationTests(TestCase):
    maxDiff = None

    @skipIf(django.VERSION < (1, 8), "extended validation not supported in Django 1.7 and below")
    def test_url_validation(self):
        v = ExtendedURLValidator()
        v('https://google.com')
        v('tel://+44(0)123-45.67#8*9')
        v('mailto://test@example.com?subject=Greetings')

        self.assertRaises(ValidationError, v, 'tel://not a phone number')
        self.assertRaises(ValidationError, v, 'mailto://not an email address')




try:
    from html.parser import HTMLParser
except ImportError:
    from HTMLParser import HTMLParser  # python 2


class InputParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)  # old-style class in Python 2
        self.data = {}

    def handle_starttag(self, tag, attributes):
        if tag == 'input':
            attributes = dict(attributes)
            name = attributes['name']
            type = attributes['type']

            try:
                if type == 'radio':
                    if 'checked' in attributes:
                        self.data[name] = attributes['value']
                elif type == 'checkbox':
                    if 'checked' in attributes:
                        self.data[name] = 'yes'
                else:
                    self.data[name] = attributes.get('value', '')
            except KeyError:
                raise RuntimeError("Faied to parse {tag}: {attributes}".format(tag=tag, attributes=attributes))

    def close(self):
        HTMLParser.close(self)  # python 2
        return self.data


def get_input_values(html):
    """
    Get the input values from a HTML snippet.

    The Python stdlib ElementTree/minidom are all XML based, and can't handle unclosed tags.
    By using HTMLParser instead of html5lib, no external dependencies are needed for testing.
    """
    parser = InputParser()
    parser.feed(html)
    return parser.close()



from __future__ import unicode_literals

import json

from django.core import serializers
from django.core.serializers.base import DeserializationError
from django.test import TestCase

from any_urlfield.models import AnyUrlValue
from any_urlfield.tests import RegPageModel, UrlModel


class SerializerTests(TestCase):

    def test_dumpdata(self):
        """
        See if the dumpdata routines handle the value properly.
        """
        page2 = RegPageModel.objects.create(slug='foo2')
        UrlModel.objects.create(
            url=AnyUrlValue.from_model(page2)
        )

        json_data = serializers.serialize("json", UrlModel.objects.all())
        data = json.loads(json_data)
        self.assertEqual(data, [{"fields": {"url": "any_urlfield.regpagemodel://1"}, "model": "any_urlfield.urlmodel", "pk": 1}])

    def test_loaddata(self):
        """
        See if the loaddata routines handle the value properly.
        """
        fixture = '[{"fields": {"url": "any_urlfield.regpagemodel://999999"}, "model": "any_urlfield.urlmodel", "pk": 9}]'

        objects = serializers.deserialize('json', fixture)
        for deserializedobject in objects:
            deserializedobject.save()

        obj = UrlModel.objects.get(pk=9)
        value = obj.url
        self.assertEqual(value.type_prefix, 'any_urlfield.regpagemodel')
        self.assertEqual(value.type_value, 999999)

    def test_loaddata_exception(self):
        """
        Passing invalid values is simply not working.
        """
        fixture = '[{"fields": {"url": "unknown.unknown://999999"}, "model": "any_urlfield.urlmodel", "pk": 9}]'

        objects = serializers.deserialize('json', fixture)
        self.assertRaises(DeserializationError, lambda: list(objects))



from __future__ import unicode_literals

from django.test import TestCase
from django.utils import six

from any_urlfield.models import AnyUrlField, AnyUrlValue
from any_urlfield.registry import UrlTypeRegistry
from any_urlfield.tests import PageModel, RegPageModel, UrlModel

unicode = six.text_type


class ModelTests(TestCase):
    maxDiff = None

    def test_registry(self):
        """
        Test the basic registry setup.
        """
        reg = UrlTypeRegistry()
        self.assertIsNotNone(reg['http'])
        self.assertIsNotNone(reg['https'])

    def test_from_model(self):
        """
        Basic test of ``from_model``.
        """
        page2 = RegPageModel.objects.create(pk=1, slug='foo2')
        v = AnyUrlValue.from_model(page2)

        self.assertEqual(v.url_type.prefix, 'any_urlfield.regpagemodel')   # app_label.modelname
        self.assertEqual(v.type_value, page2.id)
        self.assertEqual(v.to_db_value(), 'any_urlfield.regpagemodel://1')

    def test_from_db_value(self):
        """
        Basic test of ``from_db_value``.
        """
        reg = UrlTypeRegistry()

        v = AnyUrlValue.from_db_value("http://www.example.com/", reg)
        self.assertEqual(v.type_prefix, 'http')
        self.assertEqual(v.type_value, "http://www.example.com/")
        self.assertEqual(unicode(v), "http://www.example.com/")

    def test_from_db_value_https(self):
        """
        Make sure other URLs are properly serialized.
        """
        reg = UrlTypeRegistry()

        v = AnyUrlValue.from_db_value("https://www.example.com/", reg)
        self.assertEqual(v.type_prefix, 'http')   # http is the constant for external URL types
        self.assertEqual(v.type_value, "https://www.example.com/")
        self.assertEqual(unicode(v), "https://www.example.com/")

    def test_from_db_value_ftps(self):
        """
        Make sure other URLs are properly serialized.
        """
        reg = UrlTypeRegistry()

        v = AnyUrlValue.from_db_value("ftps://www.example.com/", reg)
        self.assertEqual(v.type_prefix, 'http')   # http is the constant for external URL types
        self.assertEqual(v.type_value, "ftps://www.example.com/")
        self.assertEqual(unicode(v), "ftps://www.example.com/")

    def test_from_db_value_mailto(self):
        """
        Test constructing the value from ID.
        """
        reg = UrlTypeRegistry()

        v = AnyUrlValue.from_db_value("mailto://test@example.com", reg)
        self.assertEqual(v.type_prefix, 'http')   # http is the constant for external URL types
        self.assertEqual(v.type_value, "mailto://test@example.com")
        self.assertEqual(unicode(v), "mailto://test@example.com")

    def test_valid_db_id(self):
        """
        Make sure ID values are properly stored and serialized.
        """
        reg = UrlTypeRegistry()
        urltype = reg.register(PageModel)

        page = PageModel.objects.create(slug='foo')
        v = AnyUrlValue(urltype.prefix, page.id, reg)

        # Database state
        self.assertTrue(page.id)
        self.assertEqual(urltype.prefix, 'any_urlfield.pagemodel')   # app_label.modelname
        self.assertEqual(v.type_prefix, urltype.prefix)
        self.assertEqual(v.type_value, page.id)
        self.assertEqual(v.to_db_value(), 'any_urlfield.pagemodel://1')

        # Frontend
        self.assertEqual(unicode(v), "/foo/")          # fetches model and returns get_absolute_url()

        # Programmer API's
        self.assertIs(v.get_model(), PageModel)
        self.assertEqual(v.get_object(), page)
        self.assertTrue(v.exists())

    def test_invalid_db_id(self):
        """
        Make sure invalid ID's are properly handled and recognized.
        """
        reg = UrlTypeRegistry()
        urltype = reg.register(PageModel)
        v = AnyUrlValue(urltype.prefix, 999999, reg)

        # Database state
        self.assertEqual(v.type_value, 999999)
        self.assertEqual(v.to_db_value(), 'any_urlfield.pagemodel://999999')

        # Frontend
        from any_urlfield.models.values import logger
        logger.warning("NOTE: The following statement will cause a log to output")
        self.assertEqual(unicode(v), "#DoesNotExist")       # Avoids frontend errors

        # Programmer API's
        self.assertIs(v.get_model(), PageModel)
        self.assertRaises(PageModel.DoesNotExist, lambda: v.get_object())
        self.assertFalse(v.exists())

    def test_bool_empty(self):
        """
        Make sure empty value is treated as false.
        """
        x = AnyUrlValue.from_db_value('')
        self.assertFalse(1 if x else 0)
        self.assertFalse(x.exists())

    def test_db_empty_pre_save(self):
        """
        Make sure saving empty values works.
        """
        obj = UrlModel()
        obj.save()
        self.assertTrue(obj.pk)

    def test_db_pre_save(self):
        """
        Make sure saving works
        """
        obj = UrlModel(url=AnyUrlValue.from_db_value('http://www.example.org/'))
        obj.save()
        self.assertTrue(obj.pk)

    def test_resolve_values(self):
        """
        Make sure ID values are properly stored and serialized.
        """
        reg = UrlTypeRegistry()
        urltype = reg.register(PageModel)
        page = PageModel.objects.create(slug='foo')

        valid = AnyUrlValue(urltype.prefix, page.id, reg)
        invalid = AnyUrlValue(urltype.prefix, 999999, reg)

        with self.assertNumQueries(1):
            AnyUrlValue.resolve_values([valid, invalid])
        self.assertTrue(valid._resolved_objects)
        self.assertTrue(invalid._resolved_objects)

        with self.assertNumQueries(0):
            self.assertEqual(valid.get_object(), page)
            self.assertTrue(valid.exists())
            self.assertRaises(PageModel.DoesNotExist, lambda: invalid.get_object())
            self.assertFalse(invalid.exists())

    def test_resolve_objects(self):
        """
        Make sure ID values are properly stored and serialized.
        """
        page3 = RegPageModel.objects.create(slug='foo3')
        UrlModel.objects.create(url=AnyUrlValue.from_model(page3))
        UrlModel.objects.create(url=AnyUrlValue.from_model(page3))

        qs = list(UrlModel.objects.all())
        with self.assertNumQueries(1):
            AnyUrlField.resolve_objects(qs)
        self.assertTrue(qs[0].url._resolved_objects)
        self.assertTrue(qs[1].url._resolved_objects)

        with self.assertNumQueries(0):
            for obj in qs:
                self.assertEqual(str(obj.url), '/modelform/')
                self.assertTrue(obj.url.exists())
                self.assertEqual(obj.url.get_object(), page3)



from django.conf import settings
from django.utils.translation import get_language

_ALL_LANGUAGE_CODES = [code for code, title in settings.LANGUAGES]


def get_urlfield_cache_key(model, pk, language_code=None):
    """
    The low-level function to get the cache key for a model.
    """
    return 'anyurlfield.{0}.{1}.{2}.{3}'.format(model._meta.app_label, model.__name__, pk, language_code or get_language())


def get_object_cache_keys(instance):
    """
    Return the cache keys associated with an object.
    """
    if not instance.pk or instance._state.adding:
        return []

    keys = []
    for language in _get_available_languages(instance):
        keys.append(get_urlfield_cache_key(instance.__class__, instance.pk, language))

    return keys


def _get_available_languages(instance):
    try:
        return instance.get_available_languages()  # django-parler
    except AttributeError:
        return _ALL_LANGUAGE_CODES



from __future__ import unicode_literals

from any_urlfield.forms import SimpleRawIdWidget
from any_urlfield.models import AnyUrlField
from django.contrib import admin
from django.db import models


class UrlModel(models.Model):
    """
    Example model for testing AnyUrlField
    """
    url = AnyUrlField()

    def get_absolute_url(self):
        return str(self.url)


class PageModel(models.Model):
    """
    Example model to be linking to.
    """
    slug = models.SlugField()

    def get_absolute_url(self):
        return '/{0}/'.format(self.slug)


class RegPageModel(models.Model):
    slug = models.SlugField()

    def get_absolute_url(self):
        return '/{0}/'.format(self.slug)


AnyUrlField.register_model(RegPageModel, widget=SimpleRawIdWidget(RegPageModel))
admin.site.register(RegPageModel)  # Needed for SimpleRawIdWidget to render



from __future__ import unicode_literals

import pickle

from django.test import TestCase

from any_urlfield.models import AnyUrlValue
from any_urlfield.registry import UrlTypeRegistry
from any_urlfield.tests import PageModel

try:
    from StringIO import StringIO       # Python 2
except ImportError:
    from io import BytesIO as StringIO  # Python 3


class PickleTests(TestCase):

    def test_pickle_value(self):
        """
        See if regular fields can be pickled
        """
        v1 = AnyUrlValue.from_db_value("http://www.example.com/")
        out = StringIO()
        pickle.dump(v1, out)

        # Unpickle.
        out.seek(0)
        v2 = pickle.load(out)
        self.assertEqual(v1, v2)  # Note that __eq__ is overridden for AnyUrlValue

    def test_pickle_registry(self):
        """
        Test pickle when the ``AnyUrlValue`` has a custom registry.
        """
        reg = UrlTypeRegistry()
        urltype = reg.register(PageModel)
        page = PageModel.objects.create(slug='foo')

        # See if custom registries can be pickled
        v1 = AnyUrlValue(urltype.prefix, page.id, reg)
        out = StringIO()
        pickle.dump(v1, out)

        # Unpickle.
        out.seek(0)
        v2 = pickle.load(out)
        self.assertEqual(v1, v2)  # Note that __eq__ is overridden for AnyUrlValue!



from __future__ import unicode_literals

import django
from django import forms
from django.template import Context, Template
from django.test import TestCase

import any_urlfield.forms
from any_urlfield.forms import SimpleRawIdWidget
from any_urlfield.models import AnyUrlValue
from any_urlfield.registry import UrlTypeRegistry
from any_urlfield.tests import PageModel, RegPageModel, UrlModel
from any_urlfield.tests.utils import get_input_values


class FormTests(TestCase):
    """
    Test form rendering and processing.
    """

    maxDiff = None

    def test_form_clean(self):
        """
        Basic test of form validation.
        """
        reg = UrlTypeRegistry()
        reg.register(PageModel)

        class ExampleForm(forms.Form):
            url = any_urlfield.forms.AnyUrlField(url_type_registry=reg)

        # Test 1: basic URL
        form = ExampleForm(data={
            'url_0': 'http',
            'url_1': 'http://examle.org/',
        })
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['url'], AnyUrlValue.from_db_value('http://examle.org/'))

        # Test 2: ID field
        x = PageModel.objects.create(slug='foo')
        form = ExampleForm(data={
            'url_0': 'any_urlfield.pagemodel',
            'url_1': '',
            'url_2': str(x.pk),
        })
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['url'].to_db_value(), 'any_urlfield.pagemodel://{}'.format(x.pk))
        self.assertEqual(form.cleaned_data['url'].get_object(), x)

        expected = AnyUrlValue.from_db_value('any_urlfield.pagemodel://{}'.format(x.pk), url_type_registry=reg)
        self.assertEqual(form.cleaned_data['url'], expected)

        # Test 3: invalid ID
        x = PageModel.objects.create(slug='foo')
        form = ExampleForm(data={
            'url_0': 'any_urlfield.pagemodel',
            'url_1': '',
            'url_2': '-1',
        })
        self.assertFalse(form.is_valid())

    def test_modelform(self):
        """
        Testing the model form creation
        """

        class UrlModelForm(forms.ModelForm):
            class Meta:
                model = UrlModel
                fields = ('url',)

        # Test showing the form
        instance = UrlModel.objects.create(url=AnyUrlValue.from_db_value('http://example.org/'))
        form = UrlModelForm(instance=instance)
        self.assertIsInstance(form.fields['url'], any_urlfield.forms.AnyUrlField)
        form.as_p()  # Walk through rendering code.

        # Test saving URLs
        form = UrlModelForm(instance=instance, data={
            'url_0': 'http',
            'url_1': 'http://example2.org/',
        })
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(str(instance.url), 'http://example2.org/')

        # Test saving IDs
        x = RegPageModel.objects.create(slug='modelform')
        form = UrlModelForm(instance=instance, data={
            'url_0': 'any_urlfield.regpagemodel',
            'url_2': str(x.pk),
        })
        self.assertEqual(form.errors, {})
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(str(instance.url), '/modelform/')

        # Test showing IDs
        form = UrlModelForm(instance=instance)
        assert form.initial['url'].to_db_value() == 'any_urlfield.regpagemodel://{}'.format(x.pk)
        form.as_p()  # Walk through rendering code.

    def test_render_widget(self):
        """
        See if widget rendering is consistent between Django versions
        """
        reg = UrlTypeRegistry()
        reg.register(PageModel)

        class ExampleForm(forms.Form):
            field = any_urlfield.forms.AnyUrlField(url_type_registry=reg)

        def _normalize_html(html):
            # Avoid some differences in Django versions
            html = html.replace(' checked="checked"', '')
            html = html.replace(' checked', '')
            html = html.replace(' selected="selected"', ' selected')
            html = html.replace(' required', '')
            return html

        html = Template('{{ form.field }}').render(Context({'form': ExampleForm()}))
        self.assertHTMLEqual(_normalize_html(html), _normalize_html("""
    <div class="any-urlfield-wrapper related-widget-wrapper">
      <ul class="any_urlfield-url_type radiolist inline" id="id_field_0">
        <li>
          <label for="id_field_0_0">
            <input type="radio" name="field_0" value="http"
                   class="any_urlfield-url_type radiolist inline" id="id_field_0_0"/>
            External URL</label>
        </li>
        <li>
          <label for="id_field_0_1">
            <input type="radio" name="field_0" value="any_urlfield.pagemodel"
                   class="any_urlfield-url_type radiolist inline" id="id_field_0_1"/>
            page model</label>
        </li>
      </ul>

      <p class="any_urlfield-url-http" style="clear:left">
        <input type="text" name="field_1" class="vTextField" id="id_field_1"/>
      </p>

      <p class="any_urlfield-url-any_urlfieldpagemodel" style="clear:left">
        <select name="field_2" id="id_field_2">
          <option value="" selected>---------</option>
        </select>
      </p>
    </div>
    """))

    def test_raw_id_widget(self):
        """
        Test how the raw ID widget renders.
        """
        from any_urlfield.models import AnyUrlField
        widget = AnyUrlField._static_registry['any_urlfield.regpagemodel'].get_widget()
        self.assertIsInstance(widget, SimpleRawIdWidget)

        html = widget.render(name='NAME', value="111")

        if django.VERSION >= (1, 8):
            self.assertHTMLEqual(html,
                                 '<input class="vForeignKeyRawIdAdminField" name="NAME" type="text" value="111" />'
                                 '<a href="/admin/any_urlfield/regpagemodel/?_to_field=id" class="related-lookup"'
                                 ' id="lookup_id_NAME" title="Lookup"></a>')
        else:
            self.assertHTMLEqual(html,
                                 '<input class="vForeignKeyRawIdAdminField" name="NAME" type="text" value="111" />'
                                 '<a href="/admin/any_urlfield/regpagemodel/?_to_field=id" class="related-lookup"'
                                 ' id="lookup_id_NAME" onclick="return showRelatedObjectLookupPopup(this);">'
                                 ' <img src="admin/img/selector-search.gif" width="16" height="16" alt="Lookup" /></a>')

    def test_has_changed_empty_form(self):
        """
        Test that empty placeholder forms are not considered filled in
        """
        reg = UrlTypeRegistry()
        reg.register(PageModel)

        class ExampleForm(forms.Form):
            url = any_urlfield.forms.AnyUrlField(url_type_registry=reg)

        if django.VERSION >= (1, 10):
            empty_kwargs = dict(empty_permitted=True, use_required_attribute=False)
        else:
            empty_kwargs = dict(empty_permitted=True)

        form = ExampleForm(**empty_kwargs)
        data = get_input_values(form.as_p())
        assert form.initial == {}
        assert data == {'url_0': 'http', 'url_1': ''}

        # Submit the values unchanged
        form = ExampleForm(data=data, **empty_kwargs)
        self.assertFalse(form.has_changed(), "form marked as changed!")

    def test_get_input_values(self):
        """
        Test the test code
        """
        html = (
            '<div>\n'
            '<input type="hidden" name="hdn" value="123" />\n'
            '<div>\n'
            '<input type="email" name="email" value="test@example.org" />\n'
            '<input type="checkbox" name="chk" checked="checked" />\n'
            '<input type="checkbox" name="not_chk" />\n'
            '</div>\n'
            '<input type="radio" name="rad" value="1" checked />\n'
            '<input type="radio" name="rad" value="2" />\n'
            '</div>'
        )
        self.assertEqual(get_input_values(html), {
            'hdn': "123",
            'email': "test@example.org",
            'rad': "1",
            'chk': "yes",
        })



"""
Custom data objects
"""
from __future__ import unicode_literals

import logging

from django.apps import apps
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.utils import six
from django.utils.encoding import python_2_unicode_compatible

from any_urlfield.cache import get_urlfield_cache_key

unicode = six.text_type
string_types = six.string_types

if six.PY3:
    long = int


logger = logging.getLogger('any_urlfield.models')

# Not really sure about the cache invalidation yet
URL_CACHE_TIMEOUT = 3600   # 1 hour


@python_2_unicode_compatible
class AnyUrlValue(object):
    """
    Custom value object for the :class:`~any_urlfield.models.AnyUrlField`.
    This value holds both the internal page ID, and external URL.
    It can be used to parse the database contents:

    .. code-block:: python

        value = AnyUrlValue.from_db_value(url)
        article = value.get_object()
        print unicode(value)

    A conversion to :class:`unicode` or :class:`str` causes the URL to be generated.
    This allows the field value to be used in string concatenations, or template variable evaluations:

    .. code-block:: html+django

        {{ mymodel.url }}
    """

    def __init__(self, type_prefix, type_value, url_type_registry=None):
        # Easy configuration, allowing other code to deserialize database values.
        if url_type_registry is None:
            from any_urlfield.models.fields import AnyUrlField
            url_type_registry = AnyUrlField._static_registry

        self.url_type_registry = url_type_registry
        self.url_type = url_type_registry[type_prefix]
        self.type_value = type_value
        self._resolved_objects = None
        self._url_cache = {}

        if url_type_registry.index(type_prefix) is None:
            raise ValueError("Unsupported AnyUrlValue prefix '{0}'. Supported values are: {1}".format(type_prefix, url_type_registry.keys()))

    @classmethod
    def from_model(cls, model, url_type_registry=None):
        """
        Convert a model value to this object.
        """
        # Easy configuration, allowing other code to deserialize database values.
        if url_type_registry is None:
            from any_urlfield.models.fields import AnyUrlField
            url_type_registry = AnyUrlField._static_registry

        url_type = url_type_registry.get_for_model(model.__class__)
        if url_type is None:
            raise ValueError("Unregistered model for AnyUrlValue: {0}".format(model.__class__))

        return cls(url_type.prefix, model.pk, url_type_registry)

    @classmethod
    def from_db_value(cls, url, url_type_registry=None):
        """
        Convert a serialized database value to this object.

        The value can be something like:

        * an external URL: ``http://..`` , ``https://..``
        * a custom prefix: ``customid://214``, ``customid://some/value``
        * a default "app.model" prefix: ``appname.model://31``
        """
        # Easy configuration, allowing other code to deserialize database values.
        if url_type_registry is None:
            from any_urlfield.models.fields import AnyUrlField
            url_type_registry = AnyUrlField._static_registry

        try:
            prefix, url_rest = url.split('://', 2)
        except ValueError:
            # While expecting the field to be validated,
            # Don't crash when there is old contents.
            prefix = 'http'
            url_rest = url

        url_type = url_type_registry[prefix]
        if url_type is None:
            raise ValueError("Unsupported URL prefix in database value '{0}'. Supported values are: {1}".format(url, url_type_registry.keys()))

        if url_type.has_id_value:
            if url_rest == 'None':
                return None
            id = long(url_rest)
            return AnyUrlValue(prefix, id, url_type_registry)
        else:
            return AnyUrlValue(prefix, url, url_type_registry)

    def to_db_value(self):
        """
        Convert the value into a serialized format which can be stored in the database.
        For example: ``http://www.external.url/``  or ``pageid://22``.
        """
        if self.url_type.prefix == 'http':
            return self.type_value
        elif self.type_value is None:
            return None  # avoid app.model://None
        else:
            return "{0}://{1}".format(self.url_type.prefix, self.type_value)

    def exists(self):
        """
        Check whether the references model still exists.
        """
        if self.url_type.prefix == 'http' and self.type_value:
            return True
        elif self.url_type.has_id_value:
            if self._resolved_objects is not None:
                # Test whether the in_bulk() found the object
                return self.type_value in self._resolved_objects
            else:
                Model = self.get_model()
                return Model.objects.filter(pk=self.type_value).exists()
        elif self.type_value:
            # Random other value that can't be checked
            return True
        else:
            # None or empty.
            return False

    def get_model(self):
        """
        Return the model that this value points to.
        """
        Model = self.url_type.model
        if isinstance(Model, string_types):
            app_label, model_name = Model.split(".")  # assome appname.ModelName otherwise.
            Model = apps.get_model(app_label, model_name)
        return Model

    def get_object(self):
        """
        Return the database object that the value points to.
        """
        if self.url_type.has_id_value:
            Model = self.get_model()
            if self._resolved_objects is not None:
                try:
                    return self._resolved_objects[self.type_value]
                except KeyError:
                    raise Model.DoesNotExist(
                        "No {} found with ID '{}' by resolve_objects()".format(
                            Model.__class__.__name__, self.type_value
                        )
                    )
            else:
                object = Model.objects.get(pk=self.type_value)
                self._resolved_objects = {self.type_value: object}
                return object
        else:
            return None

    @property
    def type_prefix(self):
        """
        Return the URL type prefix.
        For external URLs this is always ``"http"``.
        """
        return self.url_type.prefix

    # Python 2 support comes from python_2_unicode_compatible
    def __str__(self):
        """
        Return the URL that the value points to.
        """
        if self.url_type.has_id_value:
            if not self.type_value:
                return ""

            # First see if the URL is cached
            cache_key = get_urlfield_cache_key(self.get_model(), self.type_value)
            url = cache.get(cache_key)
            if url:
                return url

            try:
                object = self.get_object()
                url = object.get_absolute_url()
                cache.set(cache_key, url, URL_CACHE_TIMEOUT)
                return url
            except ObjectDoesNotExist as e:
                # Silently fail in templates. Avoid full page crashing.
                logger.error("Failed to generate URL for %r: %s", self, e)
                return "#{0}".format(e.__class__.__name__)
        else:
            return self.type_value or ""

    def __len__(self):
        return len(unicode(self))

    def __repr__(self):
        return str("<AnyUrlValue '{0}'>".format(self.to_db_value()))

    def __getattr__(self, item):
        return getattr(unicode(self), item)

    def __getitem__(self, item):
        return unicode(self).__getitem__(item)

    def __bool__(self):
        return bool(self.type_value)

    # Python 2 support:
    __nonzero__ = __bool__

    def __eq__(self, other):
        if not isinstance(other, AnyUrlValue):
            return NotImplemented

        return self.url_type == other.url_type \
            and self.type_value == other.type_value

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        """
        Pickle support
        """
        # Avoid pickling the registry if it's the shared one.
        from any_urlfield.models.fields import AnyUrlField
        if self.url_type_registry != AnyUrlField._static_registry:
            url_type_registry = self.url_type_registry
        else:
            url_type_registry = None

        return (url_type_registry, self.url_type.prefix, self.type_value)

    def __setstate__(self, state):
        url_type_registry, prefix, type_value = state

        from any_urlfield.models.fields import AnyUrlField
        if url_type_registry is not None:
            self.url_type_registry = url_type_registry
        else:
            self.url_type_registry = AnyUrlField._static_registry

        self.type_value = type_value
        self.url_type = self.url_type_registry[prefix]

    @classmethod
    def resolve_values(cls, values, skip_cached_urls=False):
        """
        Resolve the models for collection of AnyUrlValue objects, to avoid a query per object.
        """
        ids_to_resolve = {}
        values_by_model = {}
        for value in values:
            if value and value.url_type.has_id_value and value._resolved_objects is None:
                Model = value.url_type.model
                if skip_cached_urls and cache.get(get_urlfield_cache_key(Model, value.type_value)):
                    continue

                ids_to_resolve.setdefault(Model, set()).add(value.type_value)
                values_by_model.setdefault(Model, []).append(value)

        for Model, ids in ids_to_resolve.items():
            # When an object can't be found, it simply won't be found in the _resolved_objects dict.
            resolved_objects = Model.objects.in_bulk(ids)
            for value in values_by_model[Model]:
                value._resolved_objects = resolved_objects
